@@ -4,6 +4,8 @@
 # For full license text, see the LICENSE file in the repo root
 # or https://opensource.org/licenses/BSD-3-Clause
 
+import logging
+import os
 import subprocess
 import time
 from typing import Optional
@@ -50,12 +52,13 @@ class CUDAFunctionManager:
 
     """
 
-    def __init__(self, num_agents: int = 1, num_envs: int = 1):
+    def __init__(self, num_agents: int = 1, num_envs: int = 1, process_id: int = 0):
         """
         :param num_agents: total number of agents for each env,
             it defines the default block size
         :param num_envs: number of example_envs in parallel,
             it defines the default grid size
+        :param process_id: device ID
         """
         self._CUDA_module = None
 
@@ -64,6 +67,7 @@ class CUDAFunctionManager:
         self._cuda_function_names = []
         self._num_agents = int(num_agents)
         self._num_envs = int(num_envs)
+        self._process_id = process_id
         self._block = (self._num_agents, 1, 1)
         self._grid = (self._num_envs, 1)
         self._default_functions_initialized = False
@@ -84,7 +88,7 @@ class CUDAFunctionManager:
 
         self._CUDA_module = SourceModule(code, no_extern_c=True)
 
-        print("Successfully build and load the source code")
+        logging.info("Successfully build and load the source code")
         if default_functions_included:
             self.initialize_default_functions()
 
@@ -102,7 +106,7 @@ class CUDAFunctionManager:
         ), "CUDA module has already been loaded, not allowed to load twice"
 
         self._CUDA_module = cuda_driver.module_from_file(cubin)
-        print(f"Successfully load the cubin_file from {cubin}")
+        logging.info(f"Successfully load the cubin_file from {cubin}")
         if default_functions_included:
             self.initialize_default_functions()
 
@@ -114,10 +118,11 @@ class CUDAFunctionManager:
         template_path: Optional[str] = None,
         default_functions_included: bool = True,
         customized_env_registrar: Optional[EnvironmentRegistrar] = None,
+        event_messenger=None,
     ):
         """
         Compile a template source code, so self.num_agents and self.num_envs
-        will replace the template code at the compile time.
+        will replace the template code at compile time.
         Note: self.num_agents: total number of agents for each env,
         it defines the default block size
         self.num_envs: number of example_envs in parallel,
@@ -133,62 +138,93 @@ class CUDAFunctionManager:
         :param default_functions_included: load default function lists
         :param customized_env_registrar: CustomizedEnvironmentRegistrar object
             it provides the customized env info (e.g., source code path)for the build
+        :param event_messenger: multiprocessing Event to sync up the build
+        when using multiple processes
         """
-        # 'bin_path' is the designated cuda exe binary path that warp_drive
-        # is built into; 'header_path' is the designated cuda main source code path
-        # that warp_drive is trying to build.
-        # DO NOT CHANGE THEM!
-        bin_path = f"{get_project_root()}/warp_drive/cuda_bin"
-        header_path = f"{get_project_root()}/warp_drive/cuda_includes"
-        if template_path is None:
-            template_path = f"{get_project_root()}/warp_drive/cuda_includes"
-        update_env_header(
-            template_header_file=template_header_file,
-            path=template_path,
-            num_agents=self._num_agents,
-            num_envs=self._num_envs,
-        )
-        update_env_runner(
-            template_runner_file=template_runner_file,
-            path=template_path,
-            env_name=env_name,
-            customized_env_registrar=customized_env_registrar,
-        )
-        check_env_header(
-            header_file="env_config.h",
-            path=header_path,
-            num_envs=self.grid[0],
-            num_agents=self.block[0],
-        )
-        print(
-            f"header file {header_path}/env_config.h has num_agents: "
-            f"{self.block[0]} and num_envs: {self.grid[0]} "
-            f"that are consistent with the block and the grid"
-        )
 
-        # main_file is the source code
-        main_file = f"{header_path}/env_runner.cu"
         # cubin_file is the targeted compiled exe
+        bin_path = f"{get_project_root()}/warp_drive/cuda_bin"
         cubin_file = f"{bin_path}/env_runner.fatbin"
-        print(f"Compiling {main_file} -> {cubin_file}")
 
-        self._compile(main_file, cubin_file)
+        # Only process 0 is taking care of the compilation
+        if self._process_id > 0:
+            assert event_messenger is not None, (
+                "Event messenger is required to sync up "
+                "the compilation status among processes."
+            )
+            event_messenger.wait(timeout=12)
+
+            if not event_messenger.is_set():
+                raise Exception(
+                    f"Process {self._process_id} fails to get "
+                    f"the successful compilation message ... "
+                )
+
+        else:
+            # 'bin_path' is the designated cuda exe binary path that warp_drive
+            # is built into; 'header_path' is the designated cuda main source code path
+            # that warp_drive is trying to build.
+            # DO NOT CHANGE THEM!
+            header_path = f"{get_project_root()}/warp_drive/cuda_includes"
+            if template_path is None:
+                template_path = f"{get_project_root()}/warp_drive/cuda_includes"
+            update_env_header(
+                template_header_file=template_header_file,
+                path=template_path,
+                num_agents=self._num_agents,
+                num_envs=self._num_envs,
+            )
+            update_env_runner(
+                template_runner_file=template_runner_file,
+                path=template_path,
+                env_name=env_name,
+                customized_env_registrar=customized_env_registrar,
+            )
+            check_env_header(
+                header_file="env_config.h",
+                path=header_path,
+                num_envs=self.grid[0],
+                num_agents=self.block[0],
+            )
+            logging.debug(
+                f"header file {header_path}/env_config.h has num_agents: "
+                f"{self.block[0]} and num_envs: {self.grid[0]} "
+                f"that are consistent with the block and the grid"
+            )
+
+            # main_file is the source code
+            main_file = f"{header_path}/env_runner.cu"
+
+            logging.info(f"Compiling {main_file} -> {cubin_file}")
+
+            self._compile(main_file, cubin_file)
+
+            if event_messenger is not None:
+                event_messenger.set()
+
         self.load_cuda_from_binary_file(
             cubin=cubin_file, default_functions_included=default_functions_included
         )
 
-    def _compile(self, main_file, cubin_file):
+    @staticmethod
+    def _compile(main_file, cubin_file):
+
         bin_path = f"{get_project_root()}/warp_drive/cuda_bin"
         mkbin = f"mkdir -p {bin_path}"
+
         with subprocess.Popen(
             mkbin, shell=True, stderr=subprocess.STDOUT
         ) as mkbin_process:
             if mkbin_process.wait() != 0:
                 raise Exception("make bin file failed ... ")
-        print(f"Successfully mkdir the binary folder {bin_path}")
+        logging.info(f"Successfully mkdir the binary folder {bin_path}")
+
+        if os.path.exists(f"{cubin_file}"):
+            os.remove(f"{cubin_file}")
 
         try:
-            arch = f"{Context.get_device().compute_capability():sm_%d%d}"
+            cc = Context.get_device().compute_capability()  # compute capability
+            arch = f"sm_{cc[0]}{cc[1]}"
             cmd = f"nvcc --fatbin -arch={arch} {main_file} -o {cubin_file}"
             with subprocess.Popen(
                 cmd, shell=True, stderr=subprocess.STDOUT
@@ -200,15 +236,15 @@ class CUDAFunctionManager:
                         f"try to build the fatbin hybrid version "
                         f"of virtual PTX + gpu binary ... "
                     )
-            print(f"Running cmd: {cmd}")
-            print(
+            logging.info(f"Running cmd: {cmd}")
+            logging.info(
                 f"Successfully build the cubin_file "
                 f"from {main_file} to {cubin_file}"
             )
             return
 
         except Exception as err:
-            print(err)
+            logging.error(err)
 
         arch_codes = [
             "-code=sm_37",
@@ -236,15 +272,15 @@ class CUDAFunctionManager:
                             f"{cmd} \n"
                             f"try to build the lower gpu-code version ... "
                         )
-                print(f"Running cmd: {cmd}")
-                print(
+                logging.info(f"Running cmd: {cmd}")
+                logging.info(
                     f"Successfully build the cubin_file "
                     f"from {main_file} to {cubin_file}"
                 )
                 build_success = True
                 break
             except Exception as err:
-                print(err)
+                logging.error(err)
 
         if not build_success:
             raise Exception("build failed ... ")
@@ -270,7 +306,7 @@ class CUDAFunctionManager:
         ]
         self.initialize_functions(default_func_names)
         self._default_functions_initialized = True
-        print(
+        logging.info(
             "Successfully initialize the default CUDA functions "
             "managed by the CUDAFunctionManager"
         )
@@ -287,13 +323,13 @@ class CUDAFunctionManager:
         for fname in func_names:
             assert fname not in self._cuda_functions
             assert fname not in self._cuda_function_names
-            print(
+            logging.info(
                 f"starting to load the cuda kernel function: {fname} "
                 f"from the CUDA module "
             )
             self._cuda_functions[fname] = self._CUDA_module.get_function(fname)
             self._cuda_function_names.append(fname)
-            print(
+            logging.info(
                 f"finished loading the cuda kernel function: {fname} "
                 f"from the CUDA module, "
             )
@@ -311,7 +347,7 @@ class CUDAFunctionManager:
             cuda_driver.memcpy_htod(
                 constant_on_device, data_manager.shared_constant(cname)
             )
-            print(
+            logging.info(
                 f"Successfully initialize the CUDA shared constant {cname} "
                 f"managed by the CUDAFunctionManager"
             )
@@ -426,7 +462,7 @@ class CUDALogController:
         """
         assert (
             step > self.last_valid_step
-        ), "update_log is trying to update the exisiting timestep"
+        ), "update_log is trying to update the existing timestep"
         self._log_one_step(data_manager, step, self._env_id)
         self._update_log_mask(data_manager, step)
 
@@ -440,7 +476,7 @@ class CUDALogController:
         """
         self._env_id = env_id
         self.last_valid_step = -1
-        print(f"reset log for env {self._env_id}")
+        logging.info(f"reset log for env {self._env_id}")
         self._reset_log_mask(data_manager)
         self.update_log(data_manager, step=0)
 
@@ -507,7 +543,7 @@ class CUDALogController:
             elif "int" in dtype:
                 log_func = log_func_in_int
             else:
-                raise Exception(f"unknow dtype: {dtype}")
+                raise Exception(f"unknown dtype: {dtype}")
             log_func(
                 data_manager.device_data(f"{name}_for_log"),
                 data_manager.device_data(name),
@@ -600,7 +636,9 @@ class CUDASampler:
         free = self._function_manager.get_function("free_random")
         free(block=self._block, grid=self._grid)
         self._random_initialized = False
-        print("CUDASampler has explicitly released the random states memory in CUDA")
+        logging.info(
+            "CUDASampler has explicitly released the random states memory in CUDA"
+        )
 
     def init_random(self, seed: Optional[int] = None):
         """
@@ -609,7 +647,7 @@ class CUDASampler:
         """
         if seed is None:
             seed = time.time()
-            print(
+            logging.info(
                 f"random seed is not provided, by default, "
                 f"using the current timestamp {seed} as seed"
             )
@@ -675,9 +713,8 @@ class CUDASampler:
             grid=self._grid,
         )
 
-    def assign(
-        self, data_manager: CUDADataManager, actions: np.ndarray, action_name: str
-    ):
+    @staticmethod
+    def assign(data_manager: CUDADataManager, actions: np.ndarray, action_name: str):
         """
         Assign action to the action array directly. T
         his may be used for env testing or debugging purpose.
@@ -698,8 +735,8 @@ class CUDASampler:
 class CUDAEnvironmentReset:
     """
     CUDA Environment Reset: Manages the env reset when the game is terminated
-    inside GPU. Therefore GPU can automatically resets and
-    restarts example_envs by itself.
+    inside GPU. With this, the GPU can automatically reset and
+    restart example_envs by itself.
 
     prerequisite: CUDAFunctionManager is initialized, and the default function list
     has been successfully launched
@@ -749,7 +786,7 @@ class CUDAEnvironmentReset:
                 data_manager, mode, undo_done_after_reset
             )
         else:
-            self.reset_when_done_random(data_manager, mode, undo_done_after_reset)
+            raise NotImplementedError
 
     def reset_when_done_deterministic(
         self,
@@ -801,7 +838,7 @@ class CUDAEnvironmentReset:
                 elif "int" in dtype:
                     reset_func = self.reset_func_in_int_3d
                 else:
-                    raise Exception(f"unknow dtype: {dtype}")
+                    raise Exception(f"unknown dtype: {dtype}")
                 reset_func(
                     data_manager.device_data(name),
                     data_manager.device_data(f"{name}_at_reset"),
@@ -818,7 +855,7 @@ class CUDAEnvironmentReset:
                 elif "int" in dtype:
                     reset_func = self.reset_func_in_int_2d
                 else:
-                    raise Exception(f"unknow dtype: {dtype}")
+                    raise Exception(f"unknown dtype: {dtype}")
                 reset_func(
                     data_manager.device_data(name),
                     data_manager.device_data(f"{name}_at_reset"),
