@@ -9,6 +9,7 @@ The Trainer, PerfStats and Metrics classes
 """
 
 import json
+import logging
 import os
 import random
 import time
@@ -18,6 +19,7 @@ import torch
 import yaml
 from gym.spaces import Discrete, MultiDiscrete
 from torch import nn
+from torch.nn.parallel import DistributedDataParallel as DDP
 
 from warp_drive.managers.function_manager import CUDASampler
 from warp_drive.training.algorithms.a2c import A2C
@@ -62,6 +64,12 @@ def recursive_merge_config_dicts(config, default_config):
     return config
 
 
+def verbose_print(message, device_id=None):
+    if device_id is None:
+        device_id = 0
+    print(f"[Device {device_id}]: {message} ")
+
+
 class Trainer:
     """
     The trainer object. Contains modules train(), save_model_checkpoint() and
@@ -75,38 +83,49 @@ class Trainer:
         policy_tag_to_agent_id_map=None,
         create_separate_placeholders_for_each_policy=False,
         obs_dim_corresponding_to_num_agents="first",
+        device_id=0,
+        num_devices=1,
+        results_dir=None,
+        verbose=True,
     ):
         """
         Args:
-            env_wrapper: the wrapped environment object
-            config: the experiment run configuration
+            env_wrapper: the wrapped environment object.
+            config: the experiment run configuration.
             policy_tag_to_agent_id_map:
-                a dictionary mapping policy tag to agent ids
+                a dictionary mapping policy tag to agent ids.
             create_separate_placeholders_for_each_policy:
                 flag indicating whether there exist separate observations,
-                actions and rewards placeholders, as designed in the step
-                function. The placeholders will be used in the step() function
-                and during training.
+                actions and rewards placeholders, for each policy,
+                as designed in the step function. The placeholders will be
+                used in the step() function and during training.
                 When there's only a single policy, this flag will be False.
                 It can also be True when there are multiple policies, yet
-                all the agents have the same obs/action space, so we can
-                share the same placeholder.
-                Defaults to "False"
+                all the agents have the same obs and action space shapes,
+                so we can share the same placeholder.
+                Defaults to "False".
             obs_dim_corresponding_to_num_agents:
                 indicative of which dimension in the observation corresponds
                 to the number of agents, as designed in the step function.
                 It may be "first" or "last". In other words,
                 observations may be shaped (num_agents, *feature_dim) or
                 (*feature_dim, num_agents). This is required in order for
-                WarpDrive to process the observations correctly.
-                Defaults to "first"
+                WarpDrive to process the observations correctly. This is only
+                relevant when a single obs key corresponds to multiple agents.
+                Defaults to "first".
+            device_id: device ID
+            num_devices: number of GPU devices in the train, if more than one,
+                will use the distributed training pipeline among num_devices.
+            results_dir: (optional) name of the directory to save results into.
+            verbose:
+                if enabled, training metrics are printed to the screen.
         """
         assert env_wrapper is not None
+        assert env_wrapper.use_cuda
         assert config is not None
-        assert isinstance(policy_tag_to_agent_id_map, dict)
-        assert len(policy_tag_to_agent_id_map) > 0  # at least one policy
         assert isinstance(create_separate_placeholders_for_each_policy, bool)
         assert obs_dim_corresponding_to_num_agents in ["first", "last"]
+        self.obs_dim_corresponding_to_num_agents = obs_dim_corresponding_to_num_agents
 
         self.cuda_envs = env_wrapper
 
@@ -133,26 +152,40 @@ class Trainer:
             self.config["saving"], default_config["saving"]
         )
 
-        # Note: experiment name reflects run create time.
-        experiment_name = f"{time.time():10.0f}"
+        if results_dir is None:
+            # Use the current time as the name for the results directory.
+            results_dir = f"{time.time():10.0f}"
 
         # Directory to save model checkpoints and metrics
         self.save_dir = os.path.join(
-            self.get_config(["saving", "basedir"]),
-            self.get_config(["saving", "name"]),
-            self.get_config(["saving", "tag"]),
-            experiment_name,
+            self._get_config(["saving", "basedir"]),
+            self._get_config(["saving", "name"]),
+            self._get_config(["saving", "tag"]),
+            results_dir,
         )
         if not os.path.isdir(self.save_dir):
-            os.makedirs(self.save_dir, exist_ok=False)
+            os.makedirs(self.save_dir, exist_ok=True)
+
+        # Save the run configuration
+        config_filename = os.path.join(self.save_dir, "run_config.json")
+        with open(config_filename, "a+", encoding="utf8") as fp:
+            json.dump(self.config, fp)
+            fp.write("\n")
+
+        # Flag to determine whether to print training metrics
+        self.verbose = verbose
+
+        # Number of GPU devices in the train
+        self.num_devices = num_devices
+        self.device_id = device_id
 
         # Policies
         self.policy_tag_to_agent_id_map = policy_tag_to_agent_id_map
-        self.policies = list(self.get_config(["policy"]).keys())
+        self.policies = list(self._get_config(["policy"]).keys())
         self.policies_to_train = [
             policy
             for policy in self.policies
-            if self.get_config(["policy", policy, "to_train"])
+            if self._get_config(["policy", policy, "to_train"])
         ]
 
         # Flag indicating whether there needs to be separate placeholders / arrays
@@ -166,133 +199,29 @@ class Trainer:
             assert len(self.policies) > 1
 
         # Number of iterations algebra
-        self.num_episodes = self.get_config(["trainer", "num_episodes"])
-        self.training_batch_size = self.get_config(["trainer", "train_batch_size"])
-        self.num_envs = self.get_config(["trainer", "num_envs"])
+        self.num_episodes = self._get_config(["trainer", "num_episodes"])
+        assert self.num_episodes > 0
+        self.training_batch_size = self._get_config(["trainer", "train_batch_size"])
+        self.num_envs = self._get_config(["trainer", "num_envs"])
 
         self.training_batch_size_per_env = self.training_batch_size // self.num_envs
         assert self.training_batch_size_per_env > 0
 
-        self.block = self.cuda_envs.cuda_function_manager.block
-        self.data_manager = self.cuda_envs.cuda_data_manager
-        self.function_manager = self.cuda_envs.cuda_function_manager
-
-        # Seeding
-        seed = self.config["trainer"].get("seed", np.int32(time.time()))
-        self.cuda_sample_controller = CUDASampler(self.cuda_envs.cuda_function_manager)
-        self.cuda_sample_controller.init_random(seed)
-
-        torch.manual_seed(seed)
-        random.seed(seed)
-        np.random.seed(seed)
-
-        self.entropy_coeff = self.get_config(["trainer", "entropy_coeff"])
-        self.algorithm = self.get_config(["trainer", "algorithm"])
-        assert self.algorithm in ["A2C", "PPO"]
-        self.vf_loss_coeff = self.get_config(["trainer", "vf_loss_coeff"])
-        self.clip_grad_norm = self.get_config(["trainer", "clip_grad_norm"])
-        if self.clip_grad_norm:
-            self.max_grad_norm = self.get_config(["trainer", "max_grad_norm"])
-        self.normalize_advantage = self.get_config(["trainer", "normalize_advantage"])
-        self.normalize_return = self.get_config(["trainer", "normalize_return"])
-        if self.algorithm == "PPO":
-            self.clip_param = self.get_config(["trainer", "clip_param"])
-
-        # Define models and optimizers
-        self.models = {}
-        self.optimizers = {}
-        self.lr_schedules = {}
-        self.gammas = {}
-
-        # For logging episodic reward
-        self.num_completed_episodes = {}
-        self.episodic_reward_sum = {}
-        self.reward_running_sum = {}
-
-        # Indicates the current timestep of the policy model
-        self.current_timestep = {}
-
-        for policy in self.policies:
-            self.current_timestep[policy] = 0
-            policy_config = self.get_config(["policy", policy])
-            if policy_config["name"] == "fully_connected":
-                self.models[policy] = FullyConnected(
-                    self.cuda_envs,
-                    policy_config["model"]["fc_dims"],
-                    policy,
-                    self.policy_tag_to_agent_id_map,
-                    create_separate_placeholders_for_each_policy,
-                    obs_dim_corresponding_to_num_agents,
-                )
-            else:
-                raise NotImplementedError
-
-            # Load the model parameters (if model checkpoints are specified)
-            # Note: Loading the model checkpoint may also update the current timestep!
-            self.load_model_checkpoint(policy)
-
-            self.total_steps = self.cuda_envs.episode_length * self.num_episodes
-            self.num_iters = self.total_steps // self.training_batch_size
-
-            # Push the models to the GPU
-            self.models[policy].cuda()
-
-            # Initialize the (ADAM) optimizer
-            self.lr_schedules[policy] = ParamScheduler(policy_config["lr"])
-            initial_lr = self.lr_schedules[policy].get_param_value(
-                timestep=self.current_timestep[policy]
-            )
-            self.optimizers[policy] = torch.optim.Adam(
-                self.models[policy].parameters(), lr=initial_lr
-            )
-
-            self.gammas[policy] = policy_config["gamma"]
-
-            # Initialize episodic rewards and push to the GPU
-            num_agents_for_policy = len(self.policy_tag_to_agent_id_map[policy])
-            self.num_completed_episodes[policy] = 0
-            self.episodic_reward_sum[policy] = (
-                torch.tensor(0).type(torch.float32).cuda()
-            )
-            self.reward_running_sum[policy] = torch.zeros(
-                (self.num_envs, num_agents_for_policy)
-            ).cuda()
-
-        # Initialize the trainers
-        self.trainers = {}
-        for policy in self.policies_to_train:
-            if self.algorithm == "A2C":
-                # Advantage Actor-Critic
-                self.trainers[policy] = A2C(
-                    discount_factor_gamma=self.gammas[policy],
-                    normalize_advantage=self.normalize_advantage,
-                    normalize_return=self.normalize_return,
-                    vf_loss_coeff=self.vf_loss_coeff,
-                    entropy_coeff=self.entropy_coeff,
-                )
-                print(f"Initializing the A2C trainer for policy {policy}")
-            elif self.algorithm == "PPO":
-                # Proximal Policy Optimization
-                self.trainers[policy] = PPO(
-                    discount_factor_gamma=self.gammas[policy],
-                    clip_param=self.clip_param,
-                    normalize_advantage=self.normalize_advantage,
-                    normalize_return=self.normalize_return,
-                    vf_loss_coeff=self.vf_loss_coeff,
-                    entropy_coeff=self.entropy_coeff,
-                )
-                print(f"Initializing the PPO trainer for policy {policy}")
-            else:
-                raise NotImplementedError
+        # Push all the data and tensor arrays to the GPU
+        # upon resetting environments for the very first time.
+        self.cuda_envs.reset_all_envs()
 
         # Create and push data placeholders to the device
         create_and_push_data_placeholders(
             self.cuda_envs,
             self.policy_tag_to_agent_id_map,
-            self.training_batch_size_per_env,
             self.create_separate_placeholders_for_each_policy,
+            self.obs_dim_corresponding_to_num_agents,
+            self.training_batch_size_per_env,
         )
 
+        self.cuda_sample_controller = CUDASampler(self.cuda_envs.cuda_function_manager)
+
         # Register action placeholders
         if self.create_separate_placeholders_for_each_policy:
             for policy in self.policies:
@@ -307,20 +236,146 @@ class Trainer:
 
             self._register_actions(action_space)
 
+        # Seeding (device_id is included for distributed training)
+        seed = (
+            self.config["trainer"].get("seed", np.int32(time.time())) + self.device_id
+        )
+        self.cuda_sample_controller.init_random(seed)
+        torch.manual_seed(seed)
+        random.seed(seed)
+        np.random.seed(seed)
+
+        # Define models, optimizers, and learning rate schedules
+        self.models = {}
+        self.optimizers = {}
+        self.lr_schedules = {}
+
+        # For logging episodic reward
+        self.num_completed_episodes = {}
+        self.episodic_reward_sum = {}
+        self.reward_running_sum = {}
+
+        # Indicates the current timestep of the policy model
+        self.current_timestep = {}
+
+        self.total_steps = self.cuda_envs.episode_length * self.num_episodes
+        self.num_iters = int(self.total_steps // self.training_batch_size)
+
+        for policy in self.policies:
+            self.current_timestep[policy] = 0
+            self._initialize_policy_model(policy)
+
+        # Load the model parameters (if model checkpoints are specified)
+        # Note: Loading the model checkpoint may also update the current timestep!
+        self.load_model_checkpoint()
+
+        for policy in self.policies:
+            # Push the models to the GPU
+            self.models[policy].cuda()
+            # If distributed train, sync model using DDP
+            if self.num_devices > 1:
+                self.models[policy] = DDP(
+                    self.models[policy], device_ids=[self.device_id]
+                )
+
+            # Initialize the (ADAM) optimizer
+            lr_config = self._get_config(["policy", policy, "lr"])
+            self.lr_schedules[policy] = ParamScheduler(lr_config)
+            initial_lr = self.lr_schedules[policy].get_param_value(
+                timestep=self.current_timestep[policy]
+            )
+            self.optimizers[policy] = torch.optim.Adam(
+                self.models[policy].parameters(), lr=initial_lr
+            )
+
+            # Initialize episodic rewards and push to the GPU
+            num_agents_for_policy = len(self.policy_tag_to_agent_id_map[policy])
+            self.num_completed_episodes[policy] = 0
+            self.episodic_reward_sum[policy] = (
+                torch.tensor(0).type(torch.float32).cuda()
+            )
+            self.reward_running_sum[policy] = torch.zeros(
+                (self.num_envs, num_agents_for_policy)
+            ).cuda()
+
+        # Initialize the trainers
+        self.trainers = {}
+        self.clip_grad_norm = {}
+        self.max_grad_norm = {}
+        for policy in self.policies_to_train:
+            self._initialize_policy_algorithm(policy)
+
         # Performance Stats
         self.perf_stats = PerfStats()
 
         # Metrics
         self.metrics = Metrics()
 
-    def get_config(self, args):
+    def _initialize_policy_algorithm(self, policy):
+        algorithm = self._get_config(["policy", policy, "algorithm"])
+        assert algorithm in ["A2C", "PPO"]
+        entropy_coeff = self._get_config(["policy", policy, "entropy_coeff"])
+        vf_loss_coeff = self._get_config(["policy", policy, "vf_loss_coeff"])
+        self.clip_grad_norm[policy] = self._get_config(
+            ["policy", policy, "clip_grad_norm"]
+        )
+        if self.clip_grad_norm[policy]:
+            self.max_grad_norm[policy] = self._get_config(
+                ["policy", policy, "max_grad_norm"]
+            )
+        normalize_advantage = self._get_config(
+            ["policy", policy, "normalize_advantage"]
+        )
+        normalize_return = self._get_config(["policy", policy, "normalize_return"])
+        gamma = self._get_config(["policy", policy, "gamma"])
+        if algorithm == "A2C":
+            # Advantage Actor-Critic
+            self.trainers[policy] = A2C(
+                discount_factor_gamma=gamma,
+                normalize_advantage=normalize_advantage,
+                normalize_return=normalize_return,
+                vf_loss_coeff=vf_loss_coeff,
+                entropy_coeff=entropy_coeff,
+            )
+            logging.info(f"Initializing the A2C trainer for policy {policy}")
+        elif algorithm == "PPO":
+            # Proximal Policy Optimization
+            clip_param = self._get_config(["policy", policy, "clip_param"])
+            self.trainers[policy] = PPO(
+                discount_factor_gamma=gamma,
+                clip_param=clip_param,
+                normalize_advantage=normalize_advantage,
+                normalize_return=normalize_return,
+                vf_loss_coeff=vf_loss_coeff,
+                entropy_coeff=entropy_coeff,
+            )
+            logging.info(f"Initializing the PPO trainer for policy {policy}")
+        else:
+            raise NotImplementedError
+
+    def _initialize_policy_model(self, policy):
+        policy_model_config = self._get_config(["policy", policy, "model"])
+        if policy_model_config["type"] == "fully_connected":
+            model = FullyConnected(
+                self.cuda_envs,
+                policy_model_config["fc_dims"],
+                policy,
+                self.policy_tag_to_agent_id_map,
+                self.create_separate_placeholders_for_each_policy,
+                self.obs_dim_corresponding_to_num_agents,
+            )
+        else:
+            raise NotImplementedError
+        self.models[policy] = model
+
+    def _get_config(self, args):
         assert isinstance(args, (tuple, list))
         config = self.config
         for arg in args:
             try:
                 config = config[arg]
             except ValueError:
-                print("Missing configuration '{arg}'!")
+                logging.error("Missing configuration '{arg}'!")
         return config
 
     def train(self):
@@ -334,10 +389,10 @@ class Trainer:
             start_time = time.time()
 
             # Generate a batched rollout for every CUDA environment.
-            self.generate_rollout_batch()
+            self._generate_rollout_batch()
 
             # Train / update model parameters.
-            metrics = self.update_model_params(iteration)
+            metrics = self._update_model_params(iteration)
 
             self.perf_stats.iters = iteration + 1
             self.perf_stats.steps = self.perf_stats.iters * self.training_batch_size
@@ -345,12 +400,12 @@ class Trainer:
             self.perf_stats.total_time += end_time - start_time
 
             # Log the training metrics
-            self.log_metrics(metrics)
+            self._log_metrics(metrics)
 
             # Save torch model
             self.save_model_checkpoint(iteration)
 
-    def generate_rollout_batch(self):
+    def _generate_rollout_batch(self):
         """
         Generate an environment rollout batch.
         """
@@ -362,7 +417,7 @@ class Trainer:
 
             # Evaluate policies to compute action probabilities
             start_event.record()
-            probabilities = self.evaluate_policies(batch_index=batch_index)
+            probabilities = self._evaluate_policies(batch_index=batch_index)
             end_event.record()
             torch.cuda.synchronize()
             self.perf_stats.policy_eval_time += (
@@ -370,9 +425,9 @@ class Trainer:
             )
 
             # Sample actions using the computed probabilities
-            # and push to actions batch
+            # and push to the batch of actions
             start_event.record()
-            self.sample_actions(probabilities, batch_index=batch_index)
+            self._sample_actions(probabilities, batch_index=batch_index)
             end_event.record()
             torch.cuda.synchronize()
             self.perf_stats.action_sample_time += (
@@ -394,11 +449,12 @@ class Trainer:
             torch.cuda.synchronize()
             self.perf_stats.env_step_time += start_event.elapsed_time(end_event) / 1000
 
-    def evaluate_policies(self, batch_index):
+    def _evaluate_policies(self, batch_index=0):
         """
         Perform the policy evaluation (forward pass through the models)
         and compute action probabilities
         """
+        assert isinstance(batch_index, int)
         probabilities = {}
         for policy in self.policies:
             probabilities[policy], _ = self.models[policy](
@@ -456,7 +512,7 @@ class Trainer:
             action_dim = action_space.nvec
         else:
             raise NotImplementedError(
-                "Action spaces can be of type" "Discrete or MultiDiscrete"
+                "Only 'Discrete' or 'MultiDiscrete' type action spaces are supported!"
             )
         if len(action_dim) == 1:
             self.cuda_sample_controller.register_actions(
@@ -472,10 +528,11 @@ class Trainer:
                     num_actions=action_dim[action_idx],
                 )
 
-    def sample_actions(self, probabilities, batch_index):
+    def _sample_actions(self, probabilities, batch_index=0):
         """
         Sample action probabilities (and push the sampled actions to the device).
         """
+        assert isinstance(batch_index, int)
         if self.create_separate_placeholders_for_each_policy:
             for policy in self.policies:
                 suffix = f"_{policy}"
@@ -496,13 +553,13 @@ class Trainer:
             self.cuda_sample_controller.sample(
                 self.cuda_envs.cuda_data_manager, probabilities[0], action_name
             )
-            # Push actions to actions batch
+            # Push actions to the batch of actions
             actions = self.cuda_envs.cuda_data_manager.data_on_device_via_torch(
                 action_name
             )
             self.cuda_envs.cuda_data_manager.data_on_device_via_torch(
                 name=f"{_ACTIONS}_batch" + suffix
-            )[batch_index] = actions
+            )[batch_index, :, :, num_action_types - 1] = actions
 
         else:
             for action_idx, probs in enumerate(probabilities):
@@ -510,7 +567,7 @@ class Trainer:
                 self.cuda_sample_controller.sample(
                     self.cuda_envs.cuda_data_manager, probs, action_name
                 )
-                # Push (indexed) actions to actions and actions batch
+                # Push (indexed) actions to 'actions' and 'actions_batch'
                 actions = self.cuda_envs.cuda_data_manager.data_on_device_via_torch(
                     action_name
                 )
@@ -526,7 +583,7 @@ class Trainer:
         Push rewards and done flags to the corresponding batched versions.
         Also, update the episodic reward
         """
-
+        assert isinstance(batch_index, int)
         # Push done flags to done_flags_batch
         done_flags = (
             self.cuda_envs.cuda_data_manager.data_on_device_via_torch("_done_") > 0
@@ -581,7 +638,7 @@ class Trainer:
             # Reset the reward running sum
             self.reward_running_sum[policy][done_env_ids] = 0
 
-    def update_model_params(self, iteration):
+    def _update_model_params(self, iteration):
         start_event = torch.cuda.Event(enable_timing=True)
         end_event = torch.cuda.Event(enable_timing=True)
 
@@ -676,9 +733,9 @@ class Trainer:
             self.optimizers[policy].zero_grad()
             loss.backward()
 
-            if self.clip_grad_norm:
+            if self.clip_grad_norm[policy]:
                 nn.utils.clip_grad_norm_(
-                    self.models[policy].parameters(), self.max_grad_norm
+                    self.models[policy].parameters(), self.max_grad_norm[policy]
                 )
 
             self.optimizers[policy].step()
@@ -709,124 +766,195 @@ class Trainer:
         self.perf_stats.training_time += start_event.elapsed_time(end_event) / 1000
         return metrics_dict
 
-    def log_metrics(self, metrics):
+    def _log_metrics(self, metrics):
         # Log the metrics if it is not empty
         if len(metrics) > 0:
             perf_stats = self.perf_stats.get_perf_stats()
 
-            print("\n")
-            print("=" * 40)
-            print(
-                f"{'Iterations Completed':40}: "
-                f"{self.perf_stats.iters} / {self.num_iters}"
-            )
-            self.perf_stats.pretty_print(perf_stats)
-            self.metrics.pretty_print(metrics)
-            print("=" * 40, "\n")
+            if self.verbose:
+                print("\n")
+                print("=" * 40)
+                print(f"Device: {self.device_id}")
+                print(
+                    f"{'Iterations Completed':40}: "
+                    f"{self.perf_stats.iters} / {self.num_iters}"
+                )
+                self.perf_stats.pretty_print(perf_stats)
+                self.metrics.pretty_print(metrics)
+                print("=" * 40, "\n")
 
             # Log metrics and performance stats
             logs = {"Iterations Completed": self.perf_stats.iters}
             logs.update(metrics)
             logs.update({"Perf. Stats": perf_stats})
 
-            results_filename = os.path.join(self.save_dir, "results.json")
-            print(f"Saving the results to the file '{results_filename}'")
+            if self.num_devices > 1:
+                fn = f"results_device_{self.device_id}.json"
+            else:
+                fn = "results.json"
+            results_filename = os.path.join(self.save_dir, fn)
+            if self.verbose:
+                verbose_print(
+                    f"Saving the results to the file '{results_filename}'",
+                    self.device_id,
+                )
             with open(results_filename, "a+", encoding="utf8") as fp:
                 json.dump(logs, fp)
                 fp.write("\n")
+            self._heartbeat_check_printout(metrics)
 
-    def load_model_checkpoint(self, policy):
+    def _heartbeat_check_printout(self, metrics, check=False):
+        if check is False:
+            return
+
+        if self.num_devices > 1:
+            heartbeat_print = (
+                "Iterations Completed: "
+                + f"{self.perf_stats.iters} / {self.num_iters}: \n"
+            )
+            for policy in self.policies:
+                heartbeat_print += (
+                    f"policy '{policy}' - Mean episodic reward: "
+                    f"{metrics[policy]['Mean episodic reward']} \n"
+                )
+            verbose_print(heartbeat_print, self.device_id)
+
+    def load_model_checkpoint(self, ckpts_dict=None):
         """
         Load the model parameters if a checkpoint path is specified.
         """
-        filepath = self.config["policy"][policy]["model"]["model_ckpt_filepath"]
-        if filepath != "":
-            assert os.path.isfile(filepath), "Invalid model checkpoint path!"
-            print(
-                f"Loading the '{policy}' torch model "
-                f"from the previously saved checkpoint: '{filepath}'"
+        if ckpts_dict is None:
+            logging.info(
+                "Loading trainer model checkpoints from the run configuration."
             )
-            self.models[policy].load_state_dict(torch.load(filepath))
+            for policy in self.policies:
+                ckpt_filepath = self.config["policy"][policy]["model"][
+                    "model_ckpt_filepath"
+                ]
+                self._load_model_checkpoint_helper(policy, ckpt_filepath)
+        else:
+            assert isinstance(ckpts_dict, dict)
+            if self.verbose:
+                verbose_print(
+                    "Loading the provided trainer model checkpoints.", self.device_id
+                )
+            for policy, ckpt_filepath in ckpts_dict.items():
+                assert policy in self.policies
+                self._load_model_checkpoint_helper(policy, ckpt_filepath)
+
+    def _load_model_checkpoint_helper(self, policy, ckpt_filepath):
+        if ckpt_filepath != "":
+            assert os.path.isfile(ckpt_filepath), "Invalid model checkpoint path!"
+            if self.verbose:
+                verbose_print(
+                    f"Loading the '{policy}' torch model "
+                    f"from the previously saved checkpoint: '{ckpt_filepath}'",
+                    self.device_id,
+                )
+            self.models[policy].load_state_dict(torch.load(ckpt_filepath))
 
             # Update the current timestep using the saved checkpoint filename
-            timestep = int(filepath.split(".state_dict")[0].split("_")[-1])
-            print(f"Updating the timestep for the '{policy}' model to {timestep}.")
+            timestep = int(ckpt_filepath.split(".state_dict")[0].split("_")[-1])
+            if self.verbose:
+                verbose_print(
+                    f"Updating the timestep for the '{policy}' model to {timestep}.",
+                    self.device_id,
+                )
             self.current_timestep[policy] = timestep
 
-    def save_model_checkpoint(self, iteration):
+    def save_model_checkpoint(self, iteration=0):
         """
         Save the model parameters
         """
-        # Save model checkpoints if specified (and also for the last iteration)
-        if (
-            iteration % self.config["saving"]["model_params_save_freq"] == 0
-            or iteration == self.num_iters - 1
-        ):
-            for policy, model in self.models.items():
-                filepath = os.path.join(
-                    self.save_dir,
-                    f"{policy}_{self.current_timestep[policy]}.state_dict",
-                )
-                print(f"Saving the '{policy}' torch model to the file: '{filepath}'.")
+        # If multiple devices, save the synced-up model only for device id 0
+        if self.device_id == 0:
+            # Save model checkpoints if specified (and also for the last iteration)
+            if (
+                iteration % self.config["saving"]["model_params_save_freq"] == 0
+                or iteration == self.num_iters - 1
+            ):
+                for policy, model in self.models.items():
+                    filepath = os.path.join(
+                        self.save_dir,
+                        f"{policy}_{self.current_timestep[policy]}.state_dict",
+                    )
+                    if self.verbose:
+                        verbose_print(
+                            f"Saving the '{policy}' torch model "
+                            f"to the file: '{filepath}'.",
+                            self.device_id,
+                        )
 
-                torch.save(model.state_dict(), filepath)
+                    torch.save(model.state_dict(), filepath)
 
     def graceful_close(self):
         # Delete the sample controller to clear
         # the random seeds defined in the CUDA memory heap
         del self.cuda_sample_controller
-        print("Trainer exits gracefully")
+        if self.verbose:
+            verbose_print("Trainer exits gracefully", self.device_id)
 
-    def fetch_episode_global_states(
+    def fetch_episode_states(
         self,
+        list_of_states=None,  # list of states (data array names) to fetch
         env_id=0,  # environment id to fetch the states from
-        list_of_states=None,  # list of (global) states to fetch
     ):
         """
-        Step through env and fetch the global states for an entire episode
+        Step through env and fetch the desired states (data arrays on the GPU)
+        for an entire episode. The trained models will be used for evaluation.
         """
         assert 0 <= env_id < self.num_envs
         assert list_of_states is not None
         assert isinstance(list_of_states, list)
         assert len(list_of_states) > 0
 
+        logging.info(f"Fetching the episode states: {list_of_states} from the GPU.")
+        # Ensure env is reset before the start of training, and done flags are False
         self.cuda_envs.reset_all_envs()
         env = self.cuda_envs.env
 
-        global_states = {}
+        episode_states = {}
 
         for state in list_of_states:
-            assert self.cuda_envs.cuda_data_manager.is_data_on_device(state)
-            global_states[state] = np.zeros((env.episode_length + 1, env.num_agents))
-            global_states[state][
-                0
-            ] = self.cuda_envs.cuda_data_manager.pull_data_from_device(state)[env_id]
+            assert self.cuda_envs.cuda_data_manager.is_data_on_device(
+                state
+            ), f"{state} is not a valid array name on the GPU!"
+            # Note: Discard the first dimension, which is the env dimension
+            array_shape = self.cuda_envs.cuda_data_manager.get_shape(state)[1:]
 
-        for t in range(env.episode_length):
+            # Initialize the episode states
+            episode_states[state] = np.nan * np.stack(
+                [np.ones(array_shape) for _ in range(env.episode_length + 1)]
+            )
+
+        for timestep in range(env.episode_length + 1):
+            # Update the episode states
+            for state in list_of_states:
+                episode_states[state][
+                    timestep
+                ] = self.cuda_envs.cuda_data_manager.pull_data_from_device(state)[
+                    env_id
+                ]
             # Evaluate policies to compute action probabilities
-            probabilities = self.evaluate_policies(batch_index=t)
+            probabilities = self._evaluate_policies()
 
             # Sample actions
-            self.sample_actions(probabilities, batch_index=t)
+            self._sample_actions(probabilities)
 
             # Step through all the environments
             self.cuda_envs.step_all_envs()
 
-            # Update the global states
-            for state in list_of_states:
-                global_states[state][
-                    t + 1
-                ] = self.cuda_envs.cuda_data_manager.pull_data_from_device(state)[
-                    env_id
-                ]
-
-            # Fetch the global states when episode is complete
+            # Fetch the states when episode is complete
             if env.cuda_data_manager.pull_data_from_device("_done_")[env_id]:
+                for state in list_of_states:
+                    episode_states[state][
+                        timestep
+                    ] = self.cuda_envs.cuda_data_manager.pull_data_from_device(state)[
+                        env_id
+                    ]
                 break
 
-        return {
-            key: global_state[: t + 2] for key, global_state in global_states.items()
-        }
+        return episode_states
 
 
 class PerfStats:
@@ -861,7 +989,8 @@ class PerfStats:
             "Mean steps per sec (total)": self.steps / self.total_time,
         }
 
-    def pretty_print(self, stats):
+    @staticmethod
+    def pretty_print(stats):
         print("=" * 40)
         print("Speed performance stats")
         print("=" * 40)
